import 'dart:async';

import 'package:xmpp_stone/xmpp_stone.dart' as xmpp;

import '../../connectycube_core.dart';

import 'chat_exceptions.dart';
import 'realtime/managers/base_managers.dart';
import 'realtime/managers/chat_managers.dart';
import 'realtime/managers/chat_messages_manager.dart';
import 'realtime/managers/global_messages_manager.dart';
import 'realtime/managers/last_activity_manager.dart';
import 'realtime/managers/messages-statuses-manager.dart';
import 'realtime/managers/system_messages_manager.dart';
import 'realtime/managers/typing_statuses_manager.dart';
import 'realtime/utils/jid_utils.dart';

class CubeChatConnection {
  static final CubeChatConnection _singleton = CubeChatConnection._internal();

  CubeChatConnection._internal();

  static CubeChatConnection get instance => _singleton;

  CubeUser _currentUser;

  CubeUser get currentUser => _currentUser;

  StreamSubscription<xmpp.XmppConnectionState> _chatLoginStateSubscription;
  
  PrivateChatManager _privateChatManager;
  GroupChatManager _groupChatManager;
  ChatMessagesManager _chatMessagesManager;
  GlobalMessagesManager _globalMessagesManager;
  SystemMessagesManager _systemMessagesManager;
  MessagesStatusesManager _messagesStatusesManager;
  TypingStatusesManager _typingStatusesManager;
  LastActivityManager _lastActivityManager;

  xmpp.Connection _connection;

  xmpp.RosterManager _rosterManager;
  xmpp.MessageHandler _messageHandler;
  xmpp.PresenceManager _presenceManager;
  xmpp.VCardManager _vCardManager;

  Future<CubeUser> login(CubeUser cubeUser) async {
    Completer completer = Completer<CubeUser>();

    String resourceId = CubeSettings.instance
        .chatDefaultResource; // stone_xmpp lib doesn't support custom resource id, it is for future
    String userJid = getJidForUser(cubeUser.id, resourceId);

    xmpp.Jid jid = xmpp.Jid.fromFullJid(userJid);
    xmpp.XmppAccount account = xmpp.XmppAccount(
        userJid, jid.local, jid.domain, cubeUser.password, 5222);
    account.resource = resourceId;

    ///stone_xmpp don't support work with same connection to login/logout
    ///create new chat connection every time when login
    _connection = xmpp.Connection(account);

    _chatLoginStateSubscription =
        _connection.connectionStateStream.listen((state) {
      _onConnectionStateChangedInternal(state, completer, cubeUser);
    });

    _connection.open();

    return completer.future;
  }

  void _initCurrentUser(CubeUser cubeUser) {
    _currentUser = cubeUser;
  }

  void _initGlobalMessagesManager(xmpp.Connection connection) {
    _globalMessagesManager = globalMessagesManager;
  }

  void _initRosterManager(xmpp.Connection connection) {
    _rosterManager = xmpp.RosterManager.getInstance(connection);

//    _rosterManager.addRosterItem(xmpp.Buddy(receiverJid)).then((result) {
//      if (result.description != null) {
//        print("add roster" + result.description);
//      }
//    });
  }

  void _initPresenceManager(xmpp.Connection connection) {
    _presenceManager = xmpp.PresenceManager.getInstance(connection);
    _presenceManager.subscriptionStream.listen((streamEvent) {
//      if (streamEvent.type == xmpp.SubscriptionEventType.REQUEST) {
//        print("Accepting presence request");
//        presenceManager.acceptSubscription(streamEvent.jid);
//      }
    });
  }

  void _initVCardManager(xmpp.Connection connection) {
    _vCardManager = xmpp.VCardManager(connection);
//    _vCardManager.getSelfVCard().then((vCard) {
//      if (vCard != null) {
//        print("Your info" + vCard.buildXmlString());
//      }
//    });

//    _vCardManager.getVCardFor(receiverJid).then((vCard) {
//      if (vCard != null) {
//        print("Receiver info" + vCard.buildXmlString());
//        if (vCard != null && vCard.image != null) {
////            var file = File('test456789.jpg')
////              ..writeAsBytesSync(image.encodeJpg(vCard.image));
////            print("IMAGE SAVED TO: ${file.path}");
//        }
//      }
//    });
  }

  void _onConnectionStateChangedInternal(xmpp.XmppConnectionState state,
      Completer<CubeUser> completer, CubeUser cubeUser) {
    switch (state) {
      case xmpp.XmppConnectionState.Closed:
        log("Chat connection Closed");
        break;
      case xmpp.XmppConnectionState.ReceivingFeatures:
        log("Chat connection ReceivingFeatures");
        break;
      case xmpp.XmppConnectionState.DoneParsingFeatures:
        log("Chat connection DoneParsingFeatures");
        break;
      case xmpp.XmppConnectionState.StartTlsFailed:
        log("Chat connection StartTlsFailed");
        _notifyChatLoginError(
            completer,
            ChatConnectionException(
                message: "Open connection error: StartTlsFailed"));

        break;
      case xmpp.XmppConnectionState.AuthenticationNotSupported:
        log("Chat connection AuthenticationNotSupported");
        _notifyChatLoginError(
            completer,
            ChatConnectionException(
                message: "Open connection error: AuthenticationNotSupported"));

        break;
      case xmpp.XmppConnectionState.PlainAuthentication:
        log("Chat connection PlainAuthentication");
        break;
      case xmpp.XmppConnectionState.Authenticating:
        log("Chat connection Authenticating");
        break;
      case xmpp.XmppConnectionState.Authenticated:
        log("Chat connection Authenticated");

        _chatLoginStateSubscription?.cancel();
        _chatLoginStateSubscription = null;

        _initCurrentUser(cubeUser);

        completer.complete(cubeUser);

        break;
      case xmpp.XmppConnectionState.AuthenticationFailure:
        log("Chat connection AuthenticationFailure");
        _notifyChatLoginError(
            completer,
            ChatConnectionException(
                message: "Open connection error: AuthenticationFailure"));

        break;
      case xmpp.XmppConnectionState.SessionInitialized:
        log("Chat connection SessionInitialized");
        break;
      case xmpp.XmppConnectionState.DoneServiceDiscovery:
        log("Chat connection DoneServiceDiscovery");

        _initGlobalMessagesManager(_connection);
        _initRosterManager(_connection);
        _initPresenceManager(_connection);
        _initVCardManager(_connection);

        break;
      default:
    }
  }

  _notifyChatLoginError(
      Completer completer, Exception chatConnectionException) {
    _chatLoginStateSubscription?.cancel();
    _chatLoginStateSubscription = null;

    if (!completer.isCompleted) {
      completer.completeError(chatConnectionException);
    }
  }

  logout() {
    _chatLoginStateSubscription?.cancel();
    _chatLoginStateSubscription = null;
    _connection.close();
  }

  destroy() {
    logout();

    _connection = null;

    _currentUser = null;

    _globalMessagesManager = null;

    _rosterManager = null;
    _messageHandler = null;
    _presenceManager = null;
    _vCardManager = null;

    _closeManager(_privateChatManager);
    _closeManager(_groupChatManager);
    _closeManager(_chatMessagesManager);
    _closeManager(_systemMessagesManager);
    _closeManager(_messagesStatusesManager);
    _closeManager(_typingStatusesManager);
  }

  bool isAuthenticated() {
    return _connection != null && _connection.authenticated;
  }

  Future<int> getLasUserActivity(int userId) {
    Completer completer = Completer<int>();

    if (_lastActivityManager == null && isAuthenticated()) {
      _lastActivityManager = LastActivityManager.getInstance(_connection);
    }

    if (_lastActivityManager == null) {
      completer.completeError(ChatConnectionException(
          message:
              "Something went wrong, check login to the chat and try again"));
      return completer.future;
    }

    _lastActivityManager.getLastActivity(userId).then((seconds) {
      completer.complete(seconds);
    }).catchError((onError) => completer.completeError(onError));

    return completer.future;
  }

  get globalMessagesManager {
    if (_globalMessagesManager == null && isAuthenticated()) {
      _globalMessagesManager = GlobalMessagesManager.getInstance(_connection);
    }

    return _globalMessagesManager;
  }

  get systemMessagesManager {
    if (_systemMessagesManager == null && isAuthenticated()) {
      _systemMessagesManager = SystemMessagesManager.getInstance(_connection);
    }

    return _systemMessagesManager;
  }

  get privateChatManager {
    if (_privateChatManager == null && isAuthenticated()) {
      _privateChatManager = PrivateChatManager.getInstance(_connection);
    }

    return _privateChatManager;
  }

  get groupChatManager {
    if (_groupChatManager == null && isAuthenticated()) {
      _groupChatManager = GroupChatManager.getInstance(_connection);
    }

    return _groupChatManager;
  }

  get chatMessagesManager {
    if (_chatMessagesManager == null && isAuthenticated()) {
      _chatMessagesManager = ChatMessagesManager.getInstance(_connection);
    }

    return _chatMessagesManager;
  }

  get messagesStatusesManager {
    if (_messagesStatusesManager == null && isAuthenticated()) {
      _messagesStatusesManager =
          MessagesStatusesManager.getInstance(_connection);
    }

    return _messagesStatusesManager;
  }

  get typingStatusesManager {
    if (_typingStatusesManager == null && isAuthenticated()) {
      _typingStatusesManager = TypingStatusesManager.getInstance(_connection);
    }

    return _typingStatusesManager;
  }

  void _closeManager(StreamedManager manager) {
    manager.closeStreams();
    manager = null;
  }
}
