import 'dart:collection';
import 'dart:convert';
import 'dart:math';

import 'package:crypto/crypto.dart';

import '../models/cube_session.dart';
import '../../cube_session_manager.dart';
import '../../models/cube_settings.dart';
import '../../rest/query/query.dart';
import '../../rest/request/rest_request.dart';
import '../../users/models/cube_user.dart';
import '../../utils/consts.dart';

abstract class BaseAuthQuery<T> extends Query<T> {
  CubeUser _user;

  /// Data to login via social providers
  String _provider;
  String _accessToken;
  String _accessTokenSecret;
  String _projectId;

  BaseAuthQuery([this._user]);

  /// This constructor uses for creating query for sign in user via social networks or other providers.
  /// Different providers contains different count of parameters for authorisation and we can set they
  /// to parameter `args`.
  ///
  /// [_provider] Name of provider. All possible values can be found [CubeProvider]
  /// [args]     Parameters needed to authorisation.
  ///   [CubeProvider.FACEBOOK]
  ///                      args[0] - Access Token;
  ///   [CubeProvider.TWITTER]
  ///                      args[0] - Access Token;
  ///                      args[1] - Access Token Secret;
  ///   [CubeProvider.FIREBASE_PHONE]
  ///                      args[0] - ID of project in Firebase Console;
  ///                      args[1] - Access Token;
  ///
  ///   Please describe parameters for new Provider here.
  BaseAuthQuery.usingSocial(this._provider, List<String> args) {
    if (_provider == CubeProvider.FACEBOOK ||
        _provider == CubeProvider.TWITTER) {
      this._accessToken = args[0];
      this._accessTokenSecret = args[1];
    } else if (_provider == CubeProvider.FIREBASE_PHONE) {
      this._projectId = args[0];
      this._accessToken = args[1];
    }
  }

  @override
  setMethod(RestRequest request) {
    request.method = RequestMethod.POST;
  }

  @override
  setParams(RestRequest request) {
    Map<String, dynamic> parameters = request.params;

    if (_provider != null) {
      putValue(parameters, PROVIDER, _provider);

      if (_provider == CubeProvider.FIREBASE_PHONE) {
        putValue(parameters, FIREBASE_PHONE_TOKEN, _accessToken);
        putValue(parameters, FIREBASE_PHONE_PROJECT_ID, _projectId);
      } else {
        putValue(parameters, KEYS_TOKEN, _accessToken);

        if (_provider == CubeProvider.TWITTER) {
          putValue(parameters, KEYS_SECRET, _accessTokenSecret);
        }
      }
    } else if (_user != null) {
      putValue(parameters, USER_EMAIL, _user.email);
      putValue(parameters, USER_LOGIN, _user.login);
      putValue(parameters, USER_PASSWORD, _user.password);
    }
  }

  @override
  setBody(RestRequest request) {
    super.setBody(request);

    Map<String, dynamic> parameters = request.params;

    if (_provider != null) {
      if (_provider == CubeProvider.FIREBASE_PHONE) {
        _convertFirebaseParamsToJson(parameters);
      } else {
        _convertProviderKeysToJson(parameters);
      }
    } else if (_user != null) {
      _setupUserForBody(parameters, _user);
    }

    request.setBody(jsonEncode(parameters));
  }

  _convertFirebaseParamsToJson(Map<String, dynamic> parameters) {
    Map firebaseParamsJson = Map();

    firebaseParamsJson['access_token'] = _accessToken;
    parameters.remove(FIREBASE_PHONE_TOKEN);

    firebaseParamsJson['project_id'] = _projectId;
    parameters.remove(FIREBASE_PHONE_PROJECT_ID);

    putValue(parameters, "firebase_phone", firebaseParamsJson);
  }

  _convertProviderKeysToJson(Map<String, dynamic> parameters) {
    Map keysJson = Map();

    keysJson['token'] = _accessToken;
    parameters.remove(KEYS_TOKEN);

    if (_provider == CubeProvider.TWITTER) {
      keysJson['secret'] = _accessTokenSecret;
      parameters.remove(KEYS_SECRET);
    }

    putValue(parameters, "keys", keysJson);
  }

  _setupUserForBody(Map<String, dynamic> parameters, CubeUser user);
}

class CreateSessionQuery extends BaseAuthQuery<CubeSession> {
  CreateSessionQuery([CubeUser user]) : super(user);

  CreateSessionQuery.usingSocial(String provider, List<String> args)
      : super.usingSocial(provider, args);

  @override
  setUrl(RestRequest request) {
    request.setUrl(buildQueryUrl([AUTH_ENDPOINT]));
  }

  @override
  setParams(RestRequest request) {
    super.setParams(request);

    Map<String, dynamic> parameters = request.params;

    String appId = CubeSettings.instance.applicationId;
    String authKey = CubeSettings.instance.authorizationKey;
    int timestamp = DateTime.now().millisecondsSinceEpoch ~/ 1000;
    int nonce = Random.secure().nextInt(1 << 32);

    putValue(parameters, APP_ID, appId);
    putValue(parameters, AUTH_KEY, authKey);
    putValue(parameters, NONCE, nonce.toString());
    putValue(parameters, TIMESTAMP, timestamp.toString());

    _signRequest(request);
  }

  _signRequest(RestRequest request) {
    Map<String, dynamic> parameters = request.params;

    StringBuffer keyValueSortedString = StringBuffer();
    Map<String, String> sortedParams = SplayTreeMap.from(parameters);

    for (String key in sortedParams.keys) {
      keyValueSortedString.write("$key=${sortedParams[key]}&");
    }

    String signString = keyValueSortedString.toString();
    signString = signString.substring(0, signString.length - 1);

    String key = CubeSettings.instance.authorizationSecret;

    List<int> messageBytes = utf8.encode(signString);
    List<int> keyBytes = utf8.encode(key);

    Hmac hmac = Hmac(sha1, keyBytes);
    Digest digest = hmac.convert(messageBytes);

    String signature = digest.toString();

    putValue(parameters, SIGNATURE, signature);
  }

  @override
  CubeSession processResult(String response) {
    Map<String, dynamic> res = json.decode(response);
    CubeSession session = CubeSession.fromJson(res['session']);

    _updateActiveSession(session);

    return session;
  }

  _updateActiveSession(CubeSession session) {
    CubeSessionManager.instance.activeSession = session;
  }

  @override
  void _setupUserForBody(Map<String, dynamic> parameters, CubeUser user) {
    Map userToBody = Map();

    if (user.login != null) {
      userToBody['login'] = user.login;
      parameters.remove(USER_LOGIN);
    }

    if (user.email != null) {
      userToBody['email'] = user.email;
      parameters.remove(USER_EMAIL);
    }

    userToBody['password'] = user.password;
    parameters.remove(USER_PASSWORD);

    putValue(parameters, "user", userToBody);
  }
}

class SignInQuery extends BaseAuthQuery<CubeUser> {
  SignInQuery(CubeUser user) : super(user);

  SignInQuery.usingSocial(String provider, List<String> args)
      : super.usingSocial(provider, args);

  @override
  setUrl(RestRequest request) {
    request.setUrl(buildQueryUrl([SIGNIN_ENDPOINT]));
  }

  @override
  CubeUser processResult(String response) {
    Map<String, dynamic> res = json.decode(response);

    return CubeUser.fromJson(res['user']);
  }

  @override
  void _setupUserForBody(Map<String, dynamic> parameters, CubeUser user) {
    if (user.login != null) {
      putValue(parameters, "login", user.login);
      parameters.remove(USER_LOGIN);
    }

    if (user.email != null) {
      putValue(parameters, "email", user.email);
      parameters.remove(USER_EMAIL);
    }

    putValue(parameters, "password", user.password);
    parameters.remove(USER_PASSWORD);
  }
}
