import 'dart:async';

import 'package:intl/intl.dart';

import '../request/rest_request.dart';
import '../response/rest_response.dart';
import '../../cube_exceptions.dart';
import '../../cube_session_manager.dart';
import '../../models/cube_settings.dart';
import '../../utils/consts.dart';
import '../../utils/cube_logger.dart';

abstract class Query<T> {
  Future<T> perform() {
    Completer completer = new Completer<T>();

    try {
      RestRequest request = RestRequest();

      setupRequest(request);

      log(request.toString());

      request.perform().then((response) {
        handleResponse(response);
        completer.complete(processResult(response.getBody()));
      }).catchError((error) {
        handelError(error);
        completer.completeError(error);
      });
    } catch (e) {
      completer.completeError(e);
    }

    return completer.future;
  }

  handleResponse(RestResponse response) {
    updateTokenExpirationDate(response);
  }

  handelError(Object error) {}

  void setupRequest(RestRequest request) {
    setParams(request);
    setBody(request);
    setHeaders(request);
    setMethod(request);
    setUrl(request);

    _setApiVersion(request);
    _setFrameworkVersion(request);
    _setAuthentication(request);
  }

  void setBody(RestRequest request) {}

  void setParams(RestRequest request) {}

  void setHeaders(RestRequest request) {}

  void setMethod(RestRequest request) {}

  void setUrl(RestRequest request) {
    throw IllegalArgumentException("Request 'url' not spesified");
  }

  _setApiVersion(RestRequest request) {
    request.headers[HEADER_API_VERSION] = REST_API_VERSION;
  }

  _setFrameworkVersion(RestRequest request) {
    String versionValue = HEADER_FRAMEWORK_VERSION_VALUE_PREFIX +
        " " +
        CubeSettings.instance.versionName;
    request.headers[HEADER_FRAMEWORK_VERSION] = versionValue;
  }

  _setAuthentication(RestRequest request) {
    request.headers[HEADER_TOKEN] = CubeSessionManager.instance.getToken();
  }

  T processResult(String responseBody);

  void putValue(Map<String, dynamic> parametersMap, String key, dynamic value) {
    if (value != null) {
      parametersMap[key] = value;
    }
  }

  // Helps to build final query url, adding common url parts to specific part.
  String buildQueryUrl(List<dynamic> specificParts) {
    StringBuffer stringBuffer = StringBuffer();

    stringBuffer.write(CubeSettings.instance.apiEndpoint);

    for (dynamic part in specificParts) {
      stringBuffer.write("/");
      stringBuffer.write(part.toString());
    }

    return stringBuffer.toString();
  }

  void updateTokenExpirationDate(RestResponse response) {
    Map<String, String> headers = response.getHeaders();

    if (headers == null) return;

    if (headers.containsKey(HEADER_TOKEN_EXPIRATION_DATE.toLowerCase())) {
      DateFormat format = new DateFormat(TOKEN_EXPIRATION_DATE_FORMAT);
      DateTime expirationDate = format.parse(
          headers[HEADER_TOKEN_EXPIRATION_DATE.toLowerCase()], true);

      CubeSessionManager.instance.setTokenExpirationDate(expirationDate);
    }
  }
}
